# -*- coding: utf-8 -*-
# 14-8-8
# create by: snower

'''
MySQL asynchronous client pool.
'''

import sys
import time
from collections import deque
from . import platform
from pymysql._compat import text_type
from .client import Client
from . import log
from .util import py3


class ConnectionPoolClosedError(Exception):
    pass


class ConnectionPoolUsedError(Exception):
    pass


class ConnectionNotFoundError(Exception):
    pass


class ConnectionNotUsedError(Exception):
    pass


class ConnectionUsedError(Exception):
    pass


class WaitConnectionTimeoutError(Exception):
    pass


class Connection(Client):
    def __init__(self, pool, *args, **kwargs):
        self._pool = pool
        self.idle_time = time.time()
        self.used_time = time.time()
        super(Connection, self).__init__(*args, **kwargs)

    def close(self, remote_close=False):
        if remote_close:
            return self.do_close()
        return self._pool.release_connection(self)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    if py3:
        exec("""
async def __aenter__(self):
    return self

async def __aexit__(self, exc_type, exc_val, exc_tb):
    if self._connection.autocommit_mode:
        self.close()
    else:
        try:
            if exc_type:
                await self.rollback()
            else:
                await self.commit()
        except:
            exc_info = sys.exc_info()
            self.close(True)
            try:
                raise exc_info[1].with_traceback(exc_info[2])
            finally:
                exc_info = None
        else:
            self.close()
        """)

    def __del__(self):
        try:
            self.close()
        except ConnectionNotUsedError:
            pass

    def do_close(self):
        return super(Connection, self).close()


class RecordQueryConnection(Connection):
    def __init__(self, *args, **kwargs):
        super(RecordQueryConnection, self).__init__(*args, **kwargs)

        self._last_query_sql = ""

    def connect(self):
        future = super(RecordQueryConnection, self).connect()

        origin_query = self._connection.query
        def query(sql, unbuffered=False):
            self._last_query_sql = sql
            return origin_query(sql, unbuffered)
        self._connection.query = query

        return future

    def query(self, sql, unbuffered=False):
        self._last_query_sql = sql
        return super(RecordQueryConnection, self).query(sql, unbuffered)

    def get_last_query_sql(self):
        if isinstance(self._last_query_sql, text_type):
            return self._last_query_sql.encode("utf-8")
        return self._last_query_sql


class ConnectionPool(object):
    def __init__(self, *args, **kwargs):
        self._loop = None
        self._max_connections = kwargs.pop("max_connections") if "max_connections" in kwargs else 32
        self._idle_seconds = kwargs.pop("idle_seconds") if "idle_seconds" in kwargs else 7200
        self._wait_connection_timeout = kwargs.pop("wait_connection_timeout") if "wait_connection_timeout" in kwargs else 8
        self._debug_connection_used = kwargs.pop("debug_connection_used") if "debug_connection_used" in kwargs else False
        if self._debug_connection_used:
            self._connection_cls = RecordQueryConnection
        else:
            self._connection_cls = Connection
        self._args = args
        self._kwargs = kwargs
        self._connections = deque(maxlen = self._max_connections)
        self._used_connections = {}
        self._connections_count = 0
        self._wait_connections = deque()
        self._wait_connection_timeout_futures = deque()
        self._closed = False
        self._close_future = None
        self._check_idle_callback = False

    @property
    def closed(self):
        return self._closed

    def connection_connected_callback(self, future, connection_future):
        if (hasattr(connection_future, "_exc_info") and connection_future._exc_info is not None) \
                or (hasattr(connection_future, "_exception") and connection_future._exception is not None):
            future.set_exception(connection_future.exception())
        else:
            future.set_result(connection_future.result())

            while self._wait_connections and self._connections:
                connection = self._connections.pop()
                if connection.open:
                    if self.continue_next_wait(connection):
                        self._used_connections[id(connection)] = connection
                    else:
                        self._connections.append(connection)
                        break

            if self._wait_connections and self._connections_count - 1 < self._max_connections:
                wait_future, create_time = self._wait_connections.popleft()
                wait_time = time.time() - create_time
                if wait_time >= self._wait_connection_timeout:
                    self._loop.call_soon(wait_future.set_exception, WaitConnectionTimeoutError(
                        "Wait connection timeout, used time %.2fs." % wait_time))
                else:
                    self._loop.call_soon(self.init_connection, wait_future)

    def init_connection(self, future = None):
        self._loop = platform.current_ioloop()
        future = future or platform.Future()
        connection = self._connection_cls(self, *self._args, **self._kwargs)
        connection.set_close_callback(self.connection_close_callback)
        connection_future = connection.connect()
        self._connections_count += 1
        self._used_connections[id(connection)] = connection
        connection_future.add_done_callback(lambda connection_future: self.connection_connected_callback(future, connection_future))

        if self._idle_seconds > 0 and not self._check_idle_callback:
            self._loop.call_later(min(self._idle_seconds, 60), self.check_idle_connections)
            self._check_idle_callback = True
        return future

    def get_connection(self):
        if self._closed:
            raise ConnectionPoolClosedError("Connection pool closed.")

        while self._connections:
            connection = self._connections.pop()
            self._used_connections[id(connection)] = connection
            connection.used_time = time.time()
            if connection.open:
                future = platform.Future()
                future.set_result(connection)
                return future

        if self._connections_count < self._max_connections:
            future = self.init_connection()
        else:
            future = platform.Future()
            self._wait_connections.append((future, time.time()))
        return future

    Connection = get_connection

    connect = get_connection

    def release_connection(self, connection):
        if self._closed:
            return connection.do_close()

        if not connection.open:
            future = platform.Future()
            future.set_result(None)
            return future

        if self.continue_next_wait(connection):
            while self._wait_connections and self._connections:
                connection = self._connections.pop()
                if connection.open:
                    if self.continue_next_wait(connection):
                        self._used_connections[id(connection)] = connection
                    else:
                        self._connections.append(connection)
                        break
        else:
            try:
                del self._used_connections[id(connection)]
                self._connections.append(connection)
                connection.idle_time = time.time()
            except KeyError:
                if connection not in self._connections:
                    self._loop.call_soon(connection.do_close)
                    raise ConnectionNotFoundError("Connection not found.")
                else:
                    raise ConnectionNotUsedError("Connection is not used, you maybe close wrong connection.")

        future = platform.Future()
        future.set_result(None)
        return future

    def continue_next_wait(self, connection):
        now = time.time()
        while self._wait_connections:
            wait_future, create_time = self._wait_connections.popleft()
            wait_time = now - create_time
            if wait_time >= self._wait_connection_timeout:
                self._wait_connection_timeout_futures.append((wait_future, wait_time))
                continue
            connection.used_time = now
            self._loop.call_soon(wait_future.set_result, connection)
            if self._wait_connection_timeout_futures:
                self._loop.call_soon(self.do_wait_future_exception_timeout)
            return True

        if self._wait_connection_timeout_futures:
            self._loop.call_soon(self.do_wait_future_exception_timeout)
        return False

    def do_wait_future_exception_timeout(self):
        while self._wait_connection_timeout_futures:
            wait_future, wait_time = self._wait_connection_timeout_futures.popleft()
            self._loop.call_soon(wait_future.set_exception,
                                WaitConnectionTimeoutError("Wait connection timeout, used time %.2fs." % wait_time))

    def close_connection(self, connection):
        try:
            self._connections.remove(connection)
            self._used_connections[id(connection)] = connection
            return connection.do_close()
        except ValueError:
            raise ConnectionUsedError("Connection is used, you can not close it.")

    def connection_close_callback(self, connection):
        try:
            del self._used_connections[id(connection)]
            self._connections_count -= 1
        except KeyError:
            try:
                self._connections.remove(connection)
                self._connections_count -= 1
            except ValueError:
                log.get_log().warning("Close unknown Connection %s.", connection)
        if self._close_future and not self._used_connections and not self._connections:
            self._loop.call_soon(self._close_future.set_result, None)
            self._close_future = None

    def close(self, timeout = None):
        self._loop = platform.current_ioloop()
        if self._closed:
            raise ConnectionPoolClosedError("Connection pool closed.")
        self._closed = True

        self._close_future = close_future = platform.Future()

        if self._used_connections:
            if timeout:
                def on_timeout():
                    if self._closed and self._close_future and not self._close_future.done():
                        close_future, self._close_future = self._close_future, None
                        close_future.set_exception(
                            ConnectionPoolUsedError("Connection pool is used, you must wait all query is finish."))
                self._loop.call_later(timeout, on_timeout)

        while len(self._wait_connections):
            future, create_time = self._wait_connections.popleft()
            wait_time = time.time() - create_time
            if wait_time >= self._wait_connection_timeout:
                self._loop.call_soon(future.set_exception, WaitConnectionTimeoutError(
                    "Wait connection timeout, used time %.2fs." % wait_time))
            else:
                self._loop.call_soon(future.set_exception,
                                              ConnectionPoolClosedError("Connection pool closed."))

        while len(self._connections):
            connection = self._connections.popleft()
            self._used_connections[id(connection)] = connection
            connection.do_close()

        if not self._connections_count:
            close_future.set_result(None)
            self._close_future = None

        return close_future

    def check_idle_connections(self):
        self._loop = platform.current_ioloop()
        now = time.time()

        while self._wait_connections:
            wait_future, create_time = self._wait_connections[0]
            wait_time = now - create_time
            if wait_time < self._wait_connection_timeout:
                break
            self._wait_connections.popleft()
            self._loop.call_soon(wait_future.set_exception, WaitConnectionTimeoutError(
                "Wait connection timeout, used time %.2fs." % wait_time))

        for connection in self._used_connections.values():
            if now - connection.used_time > (self._wait_connection_timeout * 4) ** 2:
                connection.do_close()
                if self._debug_connection_used:
                    log.get_log().error("Connection used timeout close, used time %.2fs %s %s.\n%s",
                                        now - connection.used_time, connection, self, connection.get_last_query_sql())
                else:
                    log.get_log().error("Connection used timeout close, used time %.2fs %s %s.",
                                        now - connection.used_time, connection, self)
            elif now - connection.used_time > self._wait_connection_timeout ** 2 * 2:
                if self._debug_connection_used:
                    log.get_log().warning("Connection maybe not release, used time %.2fs %s %s.\n%s",
                                          now - connection.used_time, connection, self, connection.get_last_query_sql())
                else:
                    log.get_log().warning("Connection maybe not release, used time %.2fs %s %s.",
                                          now - connection.used_time, connection, self)
            elif self._debug_connection_used:
                log.get_log().warning("Connection used time %.2fs %s %s.\n%s", now - connection.used_time, connection,
                                      self, connection.get_last_query_sql())

        next_check_time = now + self._idle_seconds
        for connection in tuple(self._connections):
            if now - connection.idle_time > self._idle_seconds:
                self.close_connection(connection)
            elif connection.idle_time + self._idle_seconds < next_check_time:
                next_check_time = connection.idle_time + self._idle_seconds
                
        if not self._closed and (self._connections or self._used_connections):
            self._loop.call_later(min(next_check_time - now, 60), self.check_idle_connections)
        else:
            self._check_idle_callback = False

    def __str__(self):
        return "%s <%s,%s>" % (
        super(ConnectionPool, self).__str__(), len(self._connections), len(self._used_connections))



# -*- coding: utf-8 -*-
# 14-8-8
# create by: snower

from . import platform
from pymysql.cursors import (
    Cursor as OriginCursor, DictCursor as OriginDictCursor,
    SSCursor as OriginSSCursor, SSDictCursor as OriginSSDictCursor)
from .util import async_call_method, py3


class CursorNotReadAllDataError(Exception):
    pass


class CursorNotIterError(Exception):
    pass


class Cursor(object):
    __delegate_class__ = OriginCursor

    def __init__(self, cursor):
        self._cursor = cursor

    def __del__(self):
        if self._cursor:
            try:
                self.close()
            except: pass

    def close(self):
        if self._cursor.connection is None or not self._cursor._result or not self._cursor._result.has_next:
            self._cursor.close()
            future = platform.Future()
            future.set_result(None)
        else:
            future = async_call_method(self._cursor.close)
        return future

    def nextset(self):
        return async_call_method(self._cursor.nextset)

    def mogrify(self, query, args=None):
        return self._cursor.mogrify(query, args)

    def execute(self, query, args=None):
        return async_call_method(self._cursor.execute, query, args)

    def executemany(self, query, args):
        return async_call_method(self._cursor.executemany, query, args)

    def callproc(self, procname, args=()):
        return async_call_method(self._cursor.callproc, procname, args)

    def fetchone(self):
        return self._cursor.fetchone()

    def fetchmany(self, size=None):
        return self._cursor.fetchmany(size)

    def fetchall(self):
        return self._cursor.fetchall()

    def scroll(self, value, mode='relative'):
        return self._cursor.scroll(value, mode)

    def __iter__(self):
        return self._cursor.__iter__()

    def __getattr__(self, name):
        return getattr(self._cursor, name)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        "WARING: if cursor not read all data, the connection next query is error"
        if self._cursor._result and self._cursor._result.has_next:
            raise CursorNotReadAllDataError("If cursor not read all data, the connection next query is error.")
        self.close()

    if py3:
        exec("""
async def __aiter__(self):
    return self._cursor.__iter__()

async def __aenter__(self):
    return self

async def __aexit__(self, exc_type, exc_val, exc_tb):
    await self.close()
        """)

setattr(OriginCursor, "__tormysql_class__", Cursor)


class DictCursor(Cursor):
    __delegate_class__ = OriginDictCursor

setattr(OriginDictCursor, "__tormysql_class__", DictCursor)


class SSCursor(Cursor):
    __delegate_class__ = OriginSSCursor

    def close(self):
        if self._cursor.connection is None:
            future = platform.Future()
            future.set_result(None)
        else:
            future = async_call_method(self._cursor.close)
        return future

    def read_next(self):
        return async_call_method(self._cursor.read_next)

    def fetchone(self):
        return async_call_method(self._cursor.fetchone)

    def fetchmany(self, size=None):
        return async_call_method(self._cursor.fetchmany, size)

    def fetchall(self):
        return async_call_method(self._cursor.fetchall)

    def scroll(self, value, mode='relative'):
        return async_call_method(self._cursor.scroll, value, mode)

    def __iter__(self):
        def next():
            future = async_call_method(self._cursor.fetchone)
            if future.done() and future._result is None:
                return None
            return future
        return iter(next, None)

    def __enter__(self):
        raise AttributeError("SSCursor not support with statement")

    def __exit__(self, exc_type, exc_val, exc_tb):
        raise AttributeError("SSCursor not support with statement")

    if py3:
        exec("""
async def __aiter__(self):
    return self

async def __anext__(self):
    result = await async_call_method(self._cursor.fetchone)
    if result is None:
        raise StopAsyncIteration()
    return result

async def __aenter__(self):
    return self

async def __aexit__(self, exc_type, exc_val, exc_tb):
    await self.close()
        """)

setattr(OriginSSCursor, "__tormysql_class__", SSCursor)


class SSDictCursor(SSCursor):
    __delegate_class__ = OriginSSDictCursor

setattr(OriginSSDictCursor, "__tormysql_class__", SSDictCursor)


# -*- coding: utf-8 -*-
# 14-8-8
# create by: snower

import sys
import greenlet
from . import platform

if sys.version_info[0] >= 3:
    py3 = True
else:
    py3 = False

def async_call_method(fun, *args, **kwargs):
    future = platform.Future()
    ioloop = platform.current_ioloop()

    def finish():
        try:
            result = fun(*args, **kwargs)
            if future._callbacks:
                ioloop.call_soon(future.set_result, result)
            else:
                future.set_result(result)
        except Exception as e:
            if future._callbacks:
                ioloop.call_soon(future.set_exception, e)
            else:
                future.set_exception(e)

    child_gr = greenlet.greenlet(finish)
    child_gr.switch()

    return future



# -*- coding: utf-8 -*-
# 17/12/8
# create by: snower

from __future__ import absolute_import, division, print_function

from asyncio import coroutine, Future, events, Protocol, ensure_future

def current_ioloop():
    return events.get_event_loop()

try:
    from tornado.iostream import StreamClosedError
except ImportError:
    class StreamClosedError(IOError):
        def __init__(self, real_error=None):
            super(StreamClosedError, self).__init__('Stream is closed')
            self.real_error = real_error

class IOStream(Protocol):
    def __init__(self, address, bind_address):
        self._loop = None
        self._address = address
        self._bind_address = bind_address
        self._sock = None
        self._transport = None
        self._close_callback = None
        self._connect_future = None
        self._connect_ssl_future = None
        self._read_future = None
        self._read_bytes = 0
        self._closed = False

        self._read_buffer_size = 0
        self._read_buffer = bytearray()

    def closed(self):
        return self._closed

    def set_close_callback(self, callback):
        self._close_callback = callback

    def on_closed(self, exc_info = False):
        if self._connect_future:
            if exc_info:
                self._connect_future.set_exception(exc_info[1] if isinstance(exc_info, tuple) else exc_info)
            else:
                self._connect_future.set_exception(StreamClosedError(None))
            self._connect_future = None

        if self._connect_ssl_future:
            if exc_info:
                self._connect_ssl_future.set_exception(exc_info[1] if isinstance(exc_info, tuple) else exc_info)
            else:
                self._connect_ssl_future.set_exception(StreamClosedError(None))
            self._connect_ssl_future = None

        if self._read_future:
            if exc_info:
                self._read_future.set_exception(exc_info[1] if isinstance(exc_info, tuple) else exc_info)
            else:
                self._read_future.set_exception(StreamClosedError(None))
            self._read_future = None

        if self._close_callback:
            close_callback, self._close_callback = self._close_callback, None
            self._loop.call_soon(close_callback)

        self._closed = True

    def close(self, exc_info=False):
        if self._closed:
            return

        if self._transport:
            self._transport.close()
        else:
            self.on_closed(exc_info)

    @coroutine
    def _connect(self, address, server_hostname=None):
        if isinstance(address, (str, bytes)):
            self._transport, _ = yield from self._loop.create_unix_connection(lambda : self, address, sock=self._sock, server_hostname=server_hostname)
        else:
            self._transport, _ = yield from self._loop.create_connection(lambda : self, address[0], address[1], sock=self._sock, server_hostname=server_hostname, local_addr=self._bind_address)

    def connect(self, address, connect_timeout = 0, server_hostname = None):
        assert self._connect_future is None, 'Already connecting'

        self._loop = current_ioloop()
        future = self._connect_future = Future(loop=self._loop)
        if connect_timeout:
            def on_timeout():
                self._loop_connect_timeout = None
                if self._connect_future:
                    self.close((None, IOError("Connect timeout"), None))

            self._loop_connect_timeout = self._loop.call_later(connect_timeout, on_timeout)

        def connected(connect_future):
            if self._loop_connect_timeout:
                self._loop_connect_timeout.cancel()
                self._loop_connect_timeout = None

            if connect_future._exception is not None:
                self.on_closed(connect_future.exception())
                self._connect_future = None
            else:
                self._connect_future = None
                future.set_result(connect_future.result())

        connect_future = ensure_future(self._connect(address, server_hostname))
        connect_future.add_done_callback(connected)
        return self._connect_future

    def connection_made(self, transport):
        self._transport = transport
        if self._connect_future is None and self._connect_ssl_future is None:
            transport.close()
        else:
            self._transport.set_write_buffer_limits(1024 * 1024 * 1024)

    def data_received(self, data):
        if self._read_buffer_size:
            self._read_buffer += data
        else:
            self._read_buffer = bytearray(data)
        self._read_buffer_size += len(data)
        if self._read_future and self._read_buffer_size >= self._read_bytes:
            future, self._read_future = self._read_future, None
            self._read_buffer, data = bytearray(), self._read_buffer
            self._read_buffer_size = 0
            self._read_bytes = 0
            future.set_result(data)

    def connection_lost(self, exc):
        self.on_closed(exc)
        self._transport = None

    def eof_received(self):
        return False

    def read_bytes(self, num_bytes):
        assert self._read_future is None, "Already reading"
        if self._closed:
            raise StreamClosedError(IOError('Already Closed'))

        future = self._read_future = Future()
        self._read_bytes = num_bytes
        if self._read_buffer_size >= self._read_bytes:
            future, self._read_future = self._read_future, None
            self._read_buffer, data = bytearray(), self._read_buffer
            self._read_buffer_size = 0
            self._read_bytes = 0
            future.set_result(data)
        return future

    def write(self, data):
        if self._closed:
            raise StreamClosedError(IOError('Already Closed'))

        self._transport.write(data)

    def start_tls(self, server_side, ssl_options=None, server_hostname=None, connect_timeout=None):
        if not self._transport or self._read_future:
            raise ValueError("IOStream is not idle; cannot convert to SSL")

        self._connect_ssl_future = connect_ssl_future = Future(loop=self._loop)
        waiter = Future(loop=self._loop)

        def on_connected(future):
            if self._loop_connect_timeout:
                self._loop_connect_timeout.cancel()
                self._loop_connect_timeout = None

            if connect_ssl_future._exception is not None:
                self.on_closed(future.exception())
                self._connect_ssl_future = None
            else:
                self._connect_ssl_future = None
                connect_ssl_future.set_result(self)
        waiter.add_done_callback(on_connected)

        if connect_timeout:
            def on_timeout():
                self._loop_connect_timeout = None
                if not waiter.done():
                    self.close((None, IOError("Connect timeout"), None))

            self._loop_connect_timeout = self._loop.call_later(connect_timeout, on_timeout)

        self._transport.pause_reading()
        sock, self._transport._sock = self._transport._sock, None
        self._transport = self._loop._make_ssl_transport(
            sock, self, ssl_options, waiter,
            server_side=False, server_hostname=server_hostname)

        return connect_ssl_future

    def makefile(self, mode):
        return self


# -*- coding: utf-8 -*-
# 17/12/8
# create by: snower

from __future__ import absolute_import, division, print_function

import sys
import socket
import errno
from tornado.iostream import IOStream as BaseIOStream, SSLIOStream as BaseSSLIOStream, StreamClosedError, _ERRNO_WOULDBLOCK, ssl, ssl_wrap_socket, _client_ssl_defaults
from tornado.concurrent import Future
from tornado.gen import coroutine
from tornado.ioloop import IOLoop

if sys.version_info[0] >= 3:
    import io
    StringIO = io.BytesIO
else:
    import cStringIO
    StringIO = cStringIO.StringIO

def current_ioloop():
    return IOLoop.current()


class IOStream(BaseIOStream):
    def __init__(self, address, bind_address, socket = None, *args, **kwargs):
        if socket is None:
            socket = self.init_socket(address, bind_address)

        super(IOStream, self).__init__(socket, *args, **kwargs)

        self._write_buffer = bytearray()
        self._write_buffer_pos = 0
        self._write_buffer_size = 0

    def init_socket(self, address, bind_address):
        if not isinstance(address, tuple):
            sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        else:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
            if bind_address is not None:
                sock.bind((bind_address, 0))
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        return sock

    def connect(self, address, connect_timeout = 0, server_hostname = None):
        future = Future()
        if connect_timeout:
            def timeout():
                self._loop_connect_timeout = None
                if not self._connecting:
                    self.close((None, IOError("Connect timeout"), None))

            self._loop_connect_timeout = self.io_loop.call_later(connect_timeout, timeout)

        def connected(connect_future):
            if self._loop_connect_timeout:
                self.io_loop.remove_timeout(self._loop_connect_timeout)
                self._loop_connect_timeout = None

            if (hasattr(connect_future, "_exc_info") and connect_future._exc_info is not None) \
                    or (hasattr(connect_future, "_exception") and connect_future._exception is not None):
                future.set_exception(connect_future.exception())
            else:
                future.set_result(connect_future.result())

        connect_future = super(IOStream, self).connect(address, None, server_hostname)
        connect_future.add_done_callback(connected)
        return future

    def _handle_events(self, fd, events):
        if self._closed:
            return
        try:
            if self._connecting:
                self._handle_connect()
            if self._closed:
                return
            if events & self.io_loop.READ:
                self._handle_read()
            if self._closed:
                return
            if events & self.io_loop.WRITE:
                self._handle_write()
            if self._closed:
                return
            if events & self.io_loop.ERROR:
                self.error = self.get_fd_error()
                self.io_loop.add_callback(self.close)
                return
        except Exception:
            self.close(exc_info=True)
            raise

    def _handle_connect(self):
        super(IOStream, self)._handle_connect()

        if not self.closed():
            self._state = self.io_loop.ERROR | self.io_loop.READ
            if self._write_buffer:
                self._state = self._state | self.io_loop.WRITE
            self.io_loop.update_handler(self.fileno(), self._state)

    def _handle_read(self):
        chunk = True

        while True:
            try:
                chunk = self.socket.recv(self.read_chunk_size)
                if not chunk:
                    break
                if self._read_buffer_size:
                    self._read_buffer += chunk
                else:
                    self._read_buffer = bytearray(chunk)
                self._read_buffer_size += len(chunk)
            except (socket.error, IOError, OSError) as e:
                en = e.errno if hasattr(e, 'errno') else e.args[0]
                if en in _ERRNO_WOULDBLOCK:
                    break

                if en == errno.EINTR:
                    continue

                self.close(exc_info=True)
                return

        if self._read_future is not None and self._read_buffer_size >= self._read_bytes:
            future, self._read_future = self._read_future, None
            self._read_buffer, data = bytearray(), self._read_buffer
            self._read_buffer_size = 0
            self._read_bytes = 0
            future.set_result(data)

        if not chunk:
            self.close()
            return

    def read(self, num_bytes):
        assert self._read_future is None, "Already reading"
        if self._closed:
            raise StreamClosedError(real_error=self.error)

        future = self._read_future = Future()
        self._read_bytes = num_bytes
        self._read_partial = False
        if self._read_buffer_size >= self._read_bytes:
            future, self._read_future = self._read_future, None
            self._read_buffer, data = bytearray(), self._read_buffer
            self._read_buffer_size = 0
            self._read_bytes = 0
            future.set_result(data)
        return future

    read_bytes = read

    def _handle_write(self):
        try:
            num_bytes = self.socket.send(memoryview(self._write_buffer)[self._write_buffer_pos: self._write_buffer_pos + self._write_buffer_size])
            self._write_buffer_pos += num_bytes
            self._write_buffer_size -= num_bytes
        except (socket.error, IOError, OSError) as e:
            en = e.errno if hasattr(e, 'errno') else e.args[0]
            if en not in _ERRNO_WOULDBLOCK:
                self.close(exc_info=True)
                return

        if not self._write_buffer_size:
            if self._write_buffer_pos > 0:
                self._write_buffer = bytearray()
                self._write_buffer_pos = 0

            if self._state & self.io_loop.WRITE:
                self._state = self._state & ~self.io_loop.WRITE
                self.io_loop.update_handler(self.fileno(), self._state)

    def write(self, data):
        assert isinstance(data, (bytes, bytearray))
        if self._closed:
            raise StreamClosedError(real_error=self.error)

        if data:
            if self._write_buffer_size:
                self._write_buffer += data
            else:
                self._write_buffer = bytearray(data)
            self._write_buffer_size += len(data)

        if not self._connecting:
            self._handle_write()
            if self._write_buffer_size:
                if not self._state & self.io_loop.WRITE:
                    self._state = self._state | self.io_loop.WRITE
                    self.io_loop.update_handler(self.fileno(), self._state)

    def start_tls(self, server_side, ssl_options=None, server_hostname=None, connect_timeout = None):
        if (self._read_callback or self._read_future or
                self._write_callback or self._write_futures or
                self._connect_callback or self._connect_future or
                self._pending_callbacks or self._closed or
                self._read_buffer or self._write_buffer):
            raise ValueError("IOStream is not idle; cannot convert to SSL")

        if ssl_options is None:
            ssl_options = _client_ssl_defaults

        socket = self.socket
        self.io_loop.remove_handler(socket)
        self.socket = None
        socket = ssl_wrap_socket(socket, ssl_options,
                                 server_hostname=server_hostname,
                                 server_side=server_side,
                                 do_handshake_on_connect=False)
        orig_close_callback = self._close_callback
        self._close_callback = None

        future = Future()
        ssl_stream = SSLIOStream(socket, ssl_options=ssl_options)

        # Wrap the original close callback so we can fail our Future as well.
        # If we had an "unwrap" counterpart to this method we would need
        # to restore the original callback after our Future resolves
        # so that repeated wrap/unwrap calls don't build up layers.

        def close_callback():
            if not future.done():
                # Note that unlike most Futures returned by IOStream,
                # this one passes the underlying error through directly
                # instead of wrapping everything in a StreamClosedError
                # with a real_error attribute. This is because once the
                # connection is established it's more helpful to raise
                # the SSLError directly than to hide it behind a
                # StreamClosedError (and the client is expecting SSL
                # issues rather than network issues since this method is
                # named start_tls).
                future.set_exception(ssl_stream.error or StreamClosedError())
            if orig_close_callback is not None:
                orig_close_callback()

        if connect_timeout:
            def timeout():
                ssl_stream._loop_connect_timeout = None
                if not future.done():
                    ssl_stream.close((None, IOError("Connect timeout"), None))

            ssl_stream._loop_connect_timeout = self.io_loop.call_later(connect_timeout, timeout)

        ssl_stream.set_close_callback(close_callback)
        ssl_stream._ssl_connect_callback = lambda: future.set_result(ssl_stream)
        ssl_stream.max_buffer_size = self.max_buffer_size
        ssl_stream.read_chunk_size = self.read_chunk_size
        return future

class SSLIOStream(IOStream, BaseSSLIOStream):
    def __init__(self, socket, *args, **kwargs):
        self._ssl_options = kwargs.pop('ssl_options', _client_ssl_defaults)
        IOStream.__init__(self, None, None, socket, *args, **kwargs)

        self._ssl_accepting = True
        self._handshake_reading = False
        self._handshake_writing = False
        self._ssl_connect_callback = None
        self._loop_connect_timeout = None
        self._server_hostname = None

        # If the socket is already connected, attempt to start the handshake.
        try:
            self.socket.getpeername()
        except socket.error:
            pass
        else:
            # Indirectly start the handshake, which will run on the next
            # IOLoop iteration and then the real IO state will be set in
            # _handle_events.
            self._add_io_state(self.io_loop.WRITE)

    def _handle_read(self):
        if self._ssl_accepting:
            self._do_ssl_handshake()
            return

        chunk = True

        while True:
            try:
                chunk = self.socket.recv(self.read_chunk_size)
                if not chunk:
                    break
                if self._read_buffer_size:
                    self._read_buffer += chunk
                else:
                    self._read_buffer = bytearray(chunk)
                self._read_buffer_size += len(chunk)
            except ssl.SSLError as e:
                if e.args[0] == ssl.SSL_ERROR_WANT_READ:
                    break

                self.close(exc_info=True)
                return
            except (socket.error, IOError, OSError) as e:
                en = e.errno if hasattr(e, 'errno') else e.args[0]
                if en in _ERRNO_WOULDBLOCK:
                    break

                if en == errno.EINTR:
                    continue

                self.close(exc_info=True)
                return

        if self._read_future is not None and self._read_buffer_size >= self._read_bytes:
            future, self._read_future = self._read_future, None
            self._read_buffer, data = bytearray(), self._read_buffer
            self._read_buffer_size = 0
            self._read_bytes = 0
            future.set_result(data)

        if not chunk:
            self.close()
            return

    def _handle_write(self):
        if self._ssl_accepting:
            self._do_ssl_handshake()
            return

        try:
            num_bytes = self.socket.send(memoryview(self._write_buffer)[
                                         self._write_buffer_pos: self._write_buffer_pos + self._write_buffer_size])
            self._write_buffer_pos += num_bytes
            self._write_buffer_size -= num_bytes
        except ssl.SSLError as e:
            if e.args[0] != ssl.SSL_ERROR_WANT_WRITE:
                self.close(exc_info=True)
                return
        except (socket.error, IOError, OSError) as e:
            en = e.errno if hasattr(e, 'errno') else e.args[0]
            if en not in _ERRNO_WOULDBLOCK:
                self.close(exc_info=True)
                return

        if not self._write_buffer_size:
            if self._write_buffer_pos > 0:
                self._write_buffer = bytearray()
                self._write_buffer_pos = 0

            if self._state & self.io_loop.WRITE:
                self._state = self._state & ~self.io_loop.WRITE
                self.io_loop.update_handler(self.fileno(), self._state)

    def _run_ssl_connect_callback(self):
        if self._state & self.io_loop.WRITE:
            self._state = self._state & ~self.io_loop.WRITE
            self.io_loop.update_handler(self.fileno(), self._state)

        BaseSSLIOStream._run_ssl_connect_callback(self)

    def makefile(self, mode):
        return self